import json
from random import SystemRandom

import requests

_UNICODE_ASCII_CHARACTER_SET = ('a',
                                'b',
                                'c',
                                'd',
                                'e',
                                'f',
                                'g',
                                'h',
                                'i',
                                'j',
                                'k',
                                'l',
                                'm',
                                'n',
                                'o',
                                'p',
                                'q',
                                'r',
                                's',
                                't',
                                'u',
                                'v',
                                'w',
                                'x',
                                'y',
                                'z',
                                'A',
                                'B',
                                'C',
                                'D',
                                'E',
                                'F',
                                'G',
                                'H',
                                'I',
                                'J',
                                'K',
                                'L',
                                'M',
                                'N',
                                'O',
                                'P',
                                'Q',
                                'R',
                                'S',
                                'T',
                                'U',
                                'V',
                                'W',
                                'X',
                                'Y',
                                'Z',
                                '0',
                                '1',
                                '2',
                                '3',
                                '4',
                                '5',
                                '6',
                                '7',
                                '8',
                                '9')


async def generate_token(length=30, chars=_UNICODE_ASCII_CHARACTER_SET):
    rand = SystemRandom()
    return ''.join(rand.choice(chars) for x in range(length))


async def getToken(code: str,
             scope: list[str],
             redirect_uri: str,
             client_id: str,
             client_secret: str) -> dict[str, str]:
    url = "https://discord.com/api/oauth2/token"
    scopeStr = " ".join(scope)
    rDict: dict[str, str] = {"error": "nothing loaded"}
    data = {
        'grant_type': 'authorization_code',
        'code': code,
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_uri': redirect_uri
    }
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}
    r = requests.post(url=url,
                      data=data,
                      headers=headers)
    return r.json()  # returns the json file

async def joinUrl(clientID: str, scopes: str, redirectUri: str, state: str) -> str:
    baseUrl = "https://discord.com/oauth2/authorize?"
    urlList = [baseUrl,
               "response_type=code",
               "&",
               "client_id=",
               clientID,
               "&",
               "scope=",
               scopes,
               "&",
               "state=",
               state,
               "&",
               "redirect_uri=",
               redirectUri]
    url = "".join(urlList)
    return url

class bot_permissions:
  CREATE_INSTANT_INVITE = 0x1,
  KICK_MEMBERS = 0x2,
  BAN_MEMBERS = 0x4,
  ADMINISTRATOR = 0x8,
  MANAGE_CHANNELS = 0x10,
  MANAGE_GUILD = 0x20,
  ADD_REACTIONS = 0x40,
  VIEW_AUDIT_LOG = 0x80,
  PRIORITY_SPEAKER = 0x100,
  STREAM = 0x200,
  VIEW_CHANNEL = 0x400,
  SEND_MESSAGES = 0x800,
  SEND_TTS_MESSAGES = 0x1000,
  MANAGE_MESSAGES = 0x2000,
  EMBED_LINKS = 0x4000,
  ATTACH_FILES = 0x8000,
  READ_MESSAGE_HISTORY = 0x10000,
  MENTION_EVERYONE = 0x20000,
  USE_EXTERNAL_EMOJIS = 0x40000,
  VIEW_GUILD_INSIGHTS = 0x80000,
  CONNECT = 0x100000,
  SPEAK = 0x200000,
  MUTE_MEMBERS = 0x400000,
  DEAFEN_MEMBERS = 0x800000,
  MOVE_MEMBERS = 0x1000000,
  USE_VAD = 0x2000000,
  CHANGE_NICKNAME = 0x4000000,
  MANAGE_NICKNAMES = 0x8000000,
  MANAGE_ROLES = 0x10000000,
  MANAGE_WEBHOOKS = 0x20000000,
  MANAGE_GUILD_EXPRESSIONS = 0x40000000,
  USE_APPLICATION_COMMANDS = 0x80000000,
  REQUEST_TO_SPEAK = 0x100000000,
  MANAGE_EVENTS = 0x200000000,
  MANAGE_THREADS = 0x400000000,
  CREATE_PUBLIC_THREADS = 0x800000000,
  CREATE_PRIVATE_THREADS = 0x1000000000,
  USE_EXTERNAL_STICKER = 0x2000000000,
  SEND_MESSAGES_IN_THREADS = 0x4000000000,
  USE_EMBEDDED_ACTIVITIES = 0x8000000000,
  MODERATE_MEMBERS = 0x10000000000,
  VIEW_CREATOR_MONETIZATION_ANALYTICS = 0x20000000000,
  USE_SOUNDBOARD = 0x40000000000,
  USE_EXTERNAL_SOUNDS = 0x200000000000,
  SEND_VOICE_MESSAGES = 0x400000000000
  

async def htmlEncode(str):
    str1 = str.replace(" ", "%20")
    str2 = str1.replace(":", "%3A")
    str3 = str2.replace("/", "%2F")
    return str3



class UserObj:
    def __init__(self,
                 response) -> None:
        r = response
        self.id = r['id']
        self.username = r['username']
        self.discriminator = r['discriminator']
        if 'global_name' in r:
            self.global_name = r['global_name']
        if 'avatar' in r:
            self.avatar = r['avatar']
        if 'bot' in r:
            self.bot = r['bot']
        else:
            self.bot = False
        if 'system' in r:
            self.system = r['system']
        else:
            self.system = False
        self.mfa_enabled = r['mfa_enabled']
        self.banner = r['banner']
        self.accent_color = r['accent_color']
        self.locale = r['locale']
        if 'verified' in r:
            self.verified = r['verified']
        else:
            self.verified = None
        if 'email' in r:
            self.email = r['email']
        else:
            self.email = None
        self.flags = r['flags']
        self.premium_type = r['premium_type']
        self.public_flags = r['public_flags']
        if 'avatar_decoration' in r:
            self.avatar_decoration = r['avatar_decoration']

    def __getitem__ (self, key):
        return getattr(self, key)



from typing import List, Optional, Union

import requests

from .common import generate_token, getToken, htmlEncode, joinUrl

from .models import UserObj as uObj

apiUrl = "https://discord.com/api"


class AuthUrl:
    def __init__(self,
                 client_id: str,
                 scope: List[str],
                 redirect_uri: str) -> None:
        """
        Makes and returns a url that is used to authorize users

        :param client_id: The client ID of your discord app
        :param scope: A list of scopes you want to use
        :param redirect_uri: The redirect uri you want to use
        :type client_id: str
        :type scope: List[str]
        :type redirect_uri: str
        """
        self._client_id = client_id
        self._scope = scope
        self._redirect_uri = redirect_uri

    async def makeUrl(self) -> str:
        """
        Returns the authorization link that was made

        :async:
        :returns: The authorization link. Redirect the user to the link and after they authorize, they will return to your redirect URI.
        :rtype: str
        """
        scope = self._scope
        redirect_uri = self._redirect_uri
        client_id = self._client_id
        state = await generate_token()
        self.strScope = ""
        for s in scope:
            strScope += s
            strScope += " "
        scope1 = strScope.rstrip(" ")
        _scope = scope1.replace(" ", "%20")
        redirectUri = await htmlEncode(redirect_uri)
        url = await joinUrl(client_id, _scope, redirectUri, state)
        return url


class discordApi:

    def __init__(self,
                 client_id,
                 client_secret,
                 scope: list[str],
                 redirect_uri) -> None:
        """Where you can get an access code, use the api links to get info

        client_id - The client id of your application
        client_secret - The client secret of your application
        scope - a list of the scopes you authorized for
        redirect_uri -- the redirect_uri you want to use
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope
        self.redirect_uri = redirect_uri

    async def accessToken(self, code) -> dict[str, str]:
        """Takes values input into discordApi, and the code
        returns the response as a dictionary

        Keyword Arguments:
        code -- The code you got when the user authorized
        """
        tokenDict = getToken(code,
                             self.scope,
                             self.redirect_uri,
                             self.client_id,
                             self.client_secret)
        return await tokenDict


    class User:
        def __init__(self, access_token):
            self.access_token = access_token

        

        async def get_current_user(self):
            url = apiUrl + "/users/@me"
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'Bearer ' + self.access_token
            }
            r = requests.get(url, headers=headers)
            j = r.json()
            return uObj(j)

        async def get_user_guilds(self):
            url = apiUrl + "/users/@me/guilds"
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'Bearer ' + self.access_token
            }
            r = requests.get(url, headers=headers)
            return r.json()
            

